//-----------------------------------------------------------------------
// <copyright file="AggServer.cs" company="BME MIT">
//     Released under The Code Project Open License (CPOL) 1.02
//     Based on work of Stephan Meyn and smallguy78
//     Modified by Daniel Darvas, BME MIT
// </copyright>
//-----------------------------------------------------------------------

using System.Collections.Generic;
using System.IO;
using System.Web.UI;
using RSSLibrary;

namespace RssAggregator
{
    /// <summary>
    /// AggServer
    /// Original copyright 2004 (c) Stephan Meyn
    /// Licensed under The Code Project Open License (CPOL) 1.02
    /// Modified by BME MIT FTSRG, 2012
    /// </summary>
    public class AggServer : TinyServerLib.TinyServer
    {
        /// <summary>
        /// RSS feed store.
        /// </summary>
        private RssFeedStore store;

        /// <summary>
        /// Initializes a new instance of the AggServer class.
        /// </summary>
        /// <param name="storeToUse">The RSS feed store that the aggregator will use.</param>
        public AggServer(RssFeedStore storeToUse)
        {
            this.ServeAcceptedConnects = false;
            this.store = storeToUse;
        }

        /// <summary>
        /// Handles a GET command with the given arguments.
        /// </summary>
        /// <param name="argument">The arguments of the command.</param>
        /// <param name="parameters">The parameters of the command.</param>
        protected override void DoGet(string argument, Dictionary<string, string> parameters)
        {
            // This is the entry point of a received GET command.
            this.SendOk();

            if (argument.Contains("client-info"))
            {
                // Diagnostic function
                this.SendString(this.GetDiagnosticString(parameters));
            }
            else
            {
                // Send aggregated RSSs.
                this.SendString(
                    this.WriteLinkPage(argument.Contains("reverse")));
            }
        }

        /// <summary>
        /// Creates a string that represents the received parameters. (Diagnostic function.)
        /// </summary>
        /// <param name="parameters">The received parameters.</param>
        /// <returns>The string representation of the parameters.</returns>
        private string GetDiagnosticString(Dictionary<string, string> parameters)
        {
            System.Text.StringBuilder sb = new System.Text.StringBuilder();
            foreach (var kvp in parameters)
            {
                sb.AppendFormat("<b>{0}</b> = {1} <br>", kvp.Key, kvp.Value);
            }

            return sb.ToString();
        }

        /// <summary>
        /// Creates the webpage of aggregated feeds.
        /// </summary>
        /// <param name="reverseOrder">If it is true, the feeds will be displayed in reversed order.</param>
        /// <returns>The code of the created webpage.</returns>
        private string WriteLinkPage(bool reverseOrder = false)
        {
            // get templates 
            string feedTemplate = this.GetTemplate("feedTemplate.htm");
            string itemTemplate = this.GetTemplate("itemTemplate.htm");

            StringWriter sw = new StringWriter();
            HtmlTextWriter writer = new HtmlTextWriter(sw);
            writer.WriteFullBeginTag("html");
            writer.WriteFullBeginTag("head");
            writer.WriteFullBeginTag("title");
            writer.Write("Aggregated feeds ({0}) | RssAggregator", this.store.Urls.Count);
            writer.WriteEndTag("title");
            writer.WriteEndTag("head");
            writer.WriteFullBeginTag("body");

            List<string> urls = new List<string>(this.store.Urls);
            if (reverseOrder)
            {
                urls.Reverse();
            }

            foreach (string url in urls)
            {
                RssFeed feed = this.store[url];
                writer.Write(
                    RssReader.CreateHtml(feed, feedTemplate, itemTemplate, string.Empty, 20));
                writer.WriteLine();
            }

            writer.WriteEndTag("body");
            writer.WriteEndTag("html");
            return sw.ToString();
        }

        /// <summary>
        /// Gets the template contents.
        /// </summary>
        /// <param name="path">Path of template.</param>
        /// <returns>The content of the template.</returns>
        private string GetTemplate(string path)
        {
            string templatePath = this.GetTemplatePath(path);
            return File.ReadAllText(templatePath);
        }
    }
}
