//-----------------------------------------------------------------------
// <copyright file="Aggregator.cs" company="BME MIT">
//     Released under The Code Project Open License (CPOL) 1.02
//     Based on work of Stephan Meyn and smallguy78
//     Modified by Daniel Darvas, BME MIT
// </copyright>
//-----------------------------------------------------------------------

using System;
using System.Collections;
using System.Configuration;
using System.Threading;
using RSSLibrary;

namespace RssAggregator
{
    /// <summary>
    /// Aggregator
    /// Original copyright 2004 (c) Stephan Meyn
    /// Licensed under The Code Project Open License (CPOL) 1.02
    /// Modified by BME MIT FTSRG, 2012
    /// ---
    /// Role: aggregate feeds
    /// Responsibilities:
    /// <list type="bullet">
    /// <item>Can read app config for RSS urls</item>
    /// <item>Can download RSS items on regular basis</item>
    /// <item>Can provide RSS Summary page via web server</item>
    /// </list>
    /// </summary>
    public class Aggregator
    {
        #region Private fields
        /// <summary>
        /// The store of the RSS feeds.
        /// </summary>
        private RssFeedStore store = new RssFeedStore();

        /// <summary>
        /// Aggregator server object.
        /// </summary>
        private AggServer server;
        #endregion

        #region Constructor
        /// <summary>
        /// Initializes a new instance of the Aggregator class.
        /// </summary>
        public Aggregator()
        {
            this.MilliSecondsToSleep = 10000;
            this.server = new AggServer(this.store);
        }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the sleep interval between each check to see if feeds need updating
        /// </summary>
        public int MilliSecondsToSleep { get; set; }

        /// <summary>
        /// Gets the listening port of the server.
        /// </summary>
        public int Port 
        {
            get
            {
                return this.server.Port;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the server is running or not.
        /// </summary>
        public bool ServerRunning
        {
            get
            {
                return this.server.Running;
            }
        }
        #endregion

        #region Public methods
        /// <summary>
        /// Regularly poll the feeds and keep them upto date
        /// this method should be called in a separate thread.
        /// </summary>
        public void Run()
        {
            this.LoadConfiguration(this.server);
            try
            {
                this.server.Run();
                while (true)
                {
                    this.CheckFeeds();
                    this.server.ServeAcceptedConnects = true; // after the first "CheckFeeds" call, the connections can be served
                    Thread.Sleep(this.MilliSecondsToSleep);
                }
            }
            finally
            {
                this.server.Stop();
            }
        }

        /// <summary>
        /// Adds another RSS URL.
        /// </summary>
        /// <param name="url">The URL of the feed.</param>
        public void AddRssUrl(string url)
        {
            this.store.AddFeed(url);
        }
        #endregion

        #region private methods
        /// <summary>
        /// Check the feeds.
        /// </summary>
        private void CheckFeeds()
        {
            ArrayList forUpdate = new ArrayList();
            foreach (string url in this.store.Urls)
            {
                RssFeed feed = this.store[url];
                bool needsUpdating = false;
                if (feed == null)
                {
                    needsUpdating = true;
                }
                else
                {
                    TimeSpan age = DateTime.Now.Subtract(feed.LastDownloaded);
                    needsUpdating = age.Hours > 0;
                }

                if (needsUpdating)
                {
                    forUpdate.Add(url);
                }
            }

            foreach (string url in forUpdate)
            {
                this.UpdateFeed(url);
            }

            if (forUpdate.Count > 0)
            {
                Console.WriteLine("-- {0} feeds updated. --", forUpdate.Count);
            }
        }

        /// <summary>
        /// Update the store.
        /// </summary>
        /// <param name="url">The URL of the feed to update.</param>
        private void UpdateFeed(string url)
        {
            ////System.Diagnostics.Debug.WriteLine("Update RSSFeed", url);
            Console.WriteLine("Updating Feed: {0}", url);
            RssFeed feed = RssReader.GetFeed(url);
            this.store[url] = feed;
        }

        #endregion

        #region server

        /// <summary>
        /// Loads the configuration for the given server from app.config.
        /// </summary>
        /// <param name="server">The server to configure.</param>
        private void LoadConfiguration(AggServer server)
        {
            server.Templates = ConfigurationManager.AppSettings["TemplatePath"];
            server.DefaultPage = ConfigurationManager.AppSettings["Default"];
            server.WebRootPath = ConfigurationManager.AppSettings["WebRoot"];
            server.Port = int.Parse(ConfigurationManager.AppSettings["Port"]);
        }
        #endregion
    }
}
