//-----------------------------------------------------------------------
// <copyright file="TinyServer.cs" company="BME MIT">
//     Released under The Code Project Open License (CPOL) 1.02
//     Based on work of Stephan Meyn and smallguy78
//     Modified by Daniel Darvas, BME MIT
// </copyright>
//-----------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;

namespace TinyServerLib
{
    /// <summary>
    /// TinyServer
    /// Original copyright (c) 2004 Stephan Meyn
    /// Licensed under The Code Project Open License (CPOL) 1.02
    /// ---
    /// A tiny web server that an application can host (e.g. a news aggregator)
    /// This tiny web server only understands HTTP GET commands, nothing else - 
    /// no PUT of forms.
    /// However it is ideal for building your own specialised web server. 
    /// You can subclass TinyServer and override the function doGet() to 
    /// implement special services.
    /// </summary>
    [System.Diagnostics.CodeAnalysis.SuppressMessage
        ("Microsoft.StyleCop.CSharp.OrderingRules",
        "*",
        Justification = "This code is organized using other rules.")]
    public class TinyServer
    {
        #region constants
        /// <summary>
        /// ID string of the server.
        /// </summary>
        private const string ServerID = "TinyServer V1.1";

        #endregion

        #region private attributes
        /// <summary>
        /// Ip address of client
        /// </summary>
        private IPAddress ia;

        /// <summary>
        /// Server socket.
        /// </summary>
        private Socket serverSock;

        /// <summary>
        /// Listening port
        /// </summary>
        private int serverPort = 0;

        /// <summary>
        /// Stream for reading and writing data to the client
        /// </summary>
        private NetworkStream ns;

        /// <summary>
        /// Text reader.
        /// </summary>
        private TextReader reader;

        /// <summary>
        /// The running thread.
        /// </summary>
        private Thread runningThread = null;

        /// <summary>
        /// Hostname for current connection
        /// </summary>
        private string hostname;

        /// <summary>
        /// The state of the server.
        /// </summary>
        private bool ready = false;

        /// <summary>
        /// configured item where all templates are stored
        /// </summary>
        private string templatePath = string.Empty;

        /// <summary>
        /// Default page name.
        /// </summary>
        private string defaultPageName = "index.html";

        /// <summary>
        /// path to the web directory
        /// </summary>
        private string webRootPath = string.Empty;

        /// <summary>
        /// Stores wheter the accepted connects will be served or not.
        /// </summary>
        private bool serveAcceptedConnects = true;

        #endregion

        #region public properties
        /// <summary>
        /// Gets or sets the port the webserver will run on.
        /// </summary>
        public int Port
        {
            get
            {
                return this.serverPort;
            }

            set
            {
                this.serverPort = value;
            }
        }

        /// <summary>
        /// Gets or sets the path to the web pages.
        /// </summary>
        public string WebRootPath
        {
            get
            {
                return this.webRootPath;
            }

            set
            {
                this.webRootPath = value;
            }
        }

        /// <summary>
        /// Gets or sets the default web page name.
        /// </summary>
        public string DefaultPage
        {
            get
            {
                return this.defaultPageName;
            }

            set
            {
                this.defaultPageName = value;
            }
        }

        /// <summary>
        /// Gets or sets the path to templates used by the server.
        /// </summary>
        public string Templates
        {
            get
            {
                return this.templatePath;
            }

            set
            {
                this.templatePath = value;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the web server is running or not.
        /// </summary>
        public bool Running
        {
            get
            {
                return this.ready;
            }
        }
        #endregion

        #region protected properties
        /// <summary>
        /// Gets the name of host from last call
        /// </summary>
        protected string HostName
        {
            get
            {
                return this.hostname;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the accepted connects will be served or not.
        /// </summary>
        public bool ServeAcceptedConnects
        {
            get
            {
                return this.serveAcceptedConnects;
            }

            set
            {
                this.serveAcceptedConnects = value;
            }
        }

        /// <summary>
        /// Gets the IP address for current call
        /// </summary>
        protected IPAddress CurrentAddress
        {
            get
            {
                return this.ia;
            }
        }
        #endregion

        #region running the server
        /// <summary>
        /// Start listening to port and get HTTP calls
        /// </summary>
        public void Run()
        {
            if (this.runningThread != null)
            {
                this.runningThread.Abort();
            }

            this.runningThread = new Thread(new ThreadStart(this.RunThread));
            this.runningThread.IsBackground = true;
            this.runningThread.Start();
        }

        /// <summary>
        /// Stop web server
        /// </summary>
        public void Stop()
        {
            if (this.runningThread != null)
            {
                try
                {
                    this.runningThread.Abort();
                }
                finally
                {
                    this.runningThread = null;
                }
            }
        }

        /// <summary>
        /// Check if we can actually run. If not throw an exception.
        /// </summary>
        protected virtual void CheckCanRun()
        {
            if (this.templatePath.Length == 0)
            {
                throw new TinyServerException("Configuration Error: TemplatePath is empty");
            }

            if (this.defaultPageName.Length == 0)
            {
                throw new TinyServerException("Configuration Error: DefaultPageName not set");
            }

            if (this.webRootPath.Length == 0)
            {
                throw new TinyServerException("Configuration Error: WebRootPath not set");
            }

            if (this.serverPort == 0)
            {
                throw new TinyServerException("Configuration Error: Port not set");
            }
        }

        /// <summary>
        /// the threadstart of the web server. called by Run()
        /// </summary>
        private void RunThread()
        {
            this.CheckCanRun();
            try
            {
                this.serverSock = new Socket(
                    AddressFamily.InterNetwork,
                    SocketType.Stream,
                    ProtocolType.Tcp);

                IPEndPoint ipe = new IPEndPoint(IPAddress.Any, this.serverPort);
                this.serverSock.Bind(ipe);
                this.ready = true;
                this.Loop();
            }
            catch (Exception e)
            {
                Console.WriteLine("Exception: {0}", e.Message);
            }
            finally
            {
                this.serverSock.Close();
                this.ready = false;
            }
        }

        /// <summary>
        /// Loop listening to commands. 
        /// </summary>
        private void Loop()
        {
            while (true)
            {
                this.AcceptConnection();
                if (this.ServeAcceptedConnects)
                {
                    // If the server is not ready yet, drop the connection.
                    Dictionary<string, string> parameters;
                    string command = this.GetCommand(false, out parameters);
                    command = System.Web.HttpUtility.UrlDecode(command);
                    if (command.StartsWith("GET"))
                    {
                        this.DoGet(this.CommandArg(command), parameters);
                    }
                    else
                    {
                        this.DoUnknownCommand(command, parameters);
                    }
                }

                this.ns.Flush();
                this.ns.Close();
            }
        }
        #endregion

        #region command handlers
        /// <summary>
        /// Handle a get command. Override this method for your own implementation.
        /// </summary>
        /// <param name="argument">The argument of the get command.</param>
        /// <param name="parameters">The parameters of the command.</param>
        protected virtual void DoGet(string argument, Dictionary<string, string> parameters)
        {
            string url = GetUrl(argument);
            if (url.StartsWith("/"))
            {
                url = url.Substring(1);
            }

            if (url.Length == 0)
            {
                url = this.defaultPageName;
            }

            string path = Path.Combine(this.webRootPath, url);
            if (File.Exists(path))
            {
                this.SendOk();
                this.SendFile(path);
            }
            else
            {
                this.SendError(404, "File Not Found");
            }
        }

        /// <summary>
        /// Handle http commands other than GET.
        /// Override this function to implement extra commands.
        /// This demonstrates how to use templates
        /// </summary>
        /// <param name="command">Command text.</param>
        /// <param name="parameters">The parameters of the command.</param>
        protected virtual void DoUnknownCommand(string command, Dictionary<string, string> parameters)
        {
            if (string.IsNullOrEmpty(command) == false)
            {
                this.SendTemplate("header.html");
                this.SendTemplate("HOME.html");
                this.SendTemplate("trailer.html");
            }
        }

        /// <summary>
        /// Take an argument starting with a URL 
        /// and retrieve the URL in a form usable for System.IO.
        /// </summary>
        /// <param name="argument">The argument starting with a URL.</param>
        /// <returns>The URL part of the argument.</returns>
        protected static string GetUrl(string argument)
        {
            StringBuilder sb = new StringBuilder();
            foreach (var c in argument)
            {
                if (c == '?' || char.IsWhiteSpace(c))
                {
                    break;
                }

                sb.Append(c);
            }

            return sb.ToString();
        }

        /// <summary>
        /// Determines the arguments in a command.
        /// </summary>
        /// <param name="command">Command text.</param>
        /// <returns>array of arguments</returns>
        protected static string[] UrlArgs(string command)
        {
            int questionMarkIdx = command.IndexOf("?");
            string argsString = command.Substring(questionMarkIdx + 1);
            string[] args = argsString.Split('&');
            return args;
        }
        #endregion

        #region incoming data reception
        /// <summary>
        /// Accept a connection and return a TCPClient.
        /// </summary>
        /// <returns>The TCP client.</returns>
        private Socket AcceptConnection()
        {
            this.serverSock.Listen(1);
            Socket s = this.serverSock.Accept();
            this.ns = new NetworkStream(s, System.IO.FileAccess.ReadWrite, true); // so when to close this one?
            this.reader = new StreamReader(this.ns);
            IPEndPoint ep = (IPEndPoint)s.RemoteEndPoint;
            this.ia = ep.Address;
            this.hostname = Dns.GetHostEntry(this.ia).HostName;

            Console.WriteLine("Accepted connection from {0}", this.hostname);
            return s;
        }

        /// <summary>
        /// Get a command from a connection, return the first line
        /// and store the rest as parameters
        /// </summary>
        /// <param name="skipParams">If true, the parameters will not be processed.</param>
        /// <param name="parameters">The parameters of the received command.</param>
        /// <returns>The command.</returns>
        private string GetCommand(bool skipParams, out Dictionary<string, string> parameters)
        {
            Dictionary<string, string> localParams = null;
            if (!skipParams)
            {
                localParams = new Dictionary<string, string>();
            }

            string buf = string.Empty;
            string command = string.Empty;
            bool first = true;
            try
            {
                Console.WriteLine("---- Start Receive Command ----");
                while ((buf = this.reader.ReadLine()) != null && buf.Length > 0)
                {
                    Console.WriteLine("Received:'{0}'", buf);
                    if (first)
                    {
                        command = buf;
                        first = false;
                        if (skipParams)
                        {
                            this.reader.ReadToEnd();
                            break;
                        }
                    }
                    else
                    {
                        int colonIdx = buf.IndexOf(":");
                        if (colonIdx > 0)
                        {
                            if (!skipParams)
                            {
                                localParams.Add(buf.Substring(0, colonIdx).Trim(), buf.Substring(colonIdx + 1).Trim());
                            }
                        }
                    }
                }

                Console.WriteLine("---- End Receive Command ----");
            }
            catch (SocketException e)
            {
                Console.WriteLine("Socket exception: {0}", e.Message);
            }
            catch (IOException ioe)
            {
                Console.WriteLine("IO exception: {0}", ioe.Message);
            }

            if (skipParams)
            {
                parameters = null;
            }
            else
            {
                parameters = localParams;
            }

            return command;
        }
        #endregion

        #region response sending
        /// <summary>
        /// Sends the contents of file path through
        /// the connection on socket ns.
        /// </summary>
        /// <param name="path">Full path to file</param>
        protected void SendFile(string path)
        {
            if (!File.Exists(path))
            {
                throw new TinyServerException("Missing file: sendfile({0})", path);
            }

            ////Console.WriteLine("Sending file '{0}'", path);
            byte[] buff = new byte[2048];
            try
            {
                int d;
                using (FileStream f = File.OpenRead(path))
                {
                    while ((d = f.Read(buff, 0, 2048)) > 0)
                    {
                        this.ns.Write(buff, 0, d);
                    }
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("SendFile exception: {0}", e.Message);
            }
        }

        /// <summary>
        /// Send a template to the connection. This may contain html fragments.
        /// </summary>
        /// <param name="name">Filename of template</param>
        protected void SendTemplate(string name)
        {
            this.SendFile(this.GetTemplatePath(name));
        }

        /// <summary>
        /// Sends the contents of String name back to the browser.
        /// </summary>
        /// <param name="msg">Content of the message</param>
        protected void SendString(string msg)
        {
            byte[] buff;
            buff = System.Text.Encoding.UTF8.GetBytes(msg);

            try
            {
                this.ns.Write(buff, 0, buff.Length);
            }
            catch (Exception e)
            {
                Console.WriteLine("SendString exception: {0} ", e.Message);
            }
        }

        /// <summary>
        /// Returns an error to the web browser.
        /// </summary>
        /// <param name="errNo">Error code.</param>
        /// <param name="errString">Error text.</param>
        protected void SendError(int errNo, string errString)
        {
            this.SendString(string.Format("HTTP/1.1 {0} {1}\r\n", errNo, errString));
            this.SendString(string.Format("Date:{0}\r\n", DateTime.Now));
            this.SendString(string.Format("Server:{0}\r\n", ServerID));
            this.SendString("Content-Type: text/html; charset=utf-8\r\n");
            this.SendString("Connection: close\r\n");
        }

        /// <summary>
        /// Returns an OK message to the browser.
        /// </summary>
        protected void SendOk()
        {
            this.SendString("HTTP/1.1 200 OK\r\n");
            this.SendString(string.Format("Date:{0}\r\n", DateTime.Now));
            this.SendString(string.Format("Server:{0}\r\n", ServerID));
            this.SendString("Content-Type: text/html; charset=utf-8\r\n\r\n");
        }
        #endregion

        #region utility
        /// <summary>
        /// Get the path for the required template.
        /// </summary>
        /// <param name="name">File name of the template.</param>
        /// <returns>The full path of the template.</returns>
        protected string GetTemplatePath(string name)
        {
            return Path.Combine(this.templatePath, name);
        }

        /// <summary>
        /// Returns the command argument part of an HTTP command.
        /// That is the bit after the GET. That include a URL 
        /// and optional arguments
        /// </summary>
        /// <param name="command">Command text.</param>
        /// <returns>Argument part of the command.</returns>
        protected virtual string CommandArg(string command)
        {
            int idx = 0;

            // skip blanks before command
            while ((idx < command.Length) && char.IsWhiteSpace(command, idx))
            {
                idx++;
            }

            if (idx >= command.Length)
            {
                return string.Empty;
            }

            // skip command
            while ((idx < command.Length) && !char.IsWhiteSpace(command, idx))
            {
                idx++;
            }

            if (idx >= command.Length)
            {
                return string.Empty;
            }

            return command.Substring(idx).Trim();
        }
        #endregion
    }
}
